from bppy.utils.dfs import DFSBProgram
from bppy.utils.exceptions import BPAssertionError


class DFSBProgramVerifier:
    """
    A class to verify a behavioral program using Depth First Search (DFS). Currently, this class verifies only
    properties which can be defined using assertions in b-threads. For additional properties, or programs with large
    state space, use :class:`SymbolicBProgramVerifier
    <bppy.analysis.symbolic_bprogram_verifier.SymbolicBProgramVerifier>`.

    Attributes:
    -----------
    bprogram_generator : function
        A function that generates a new instance of the BProgram.
    max_trace_length : int
        Maximum length of the trace before terminating the search.
    """
    def __init__(self, bprogram_generator, max_trace_length=1000):
        """
        Initialize a DFSBProgramVerifier.

        Parameters:
        -----------
        bprogram_generator : function
            A function that generates a new instance of the BProgram.
        max_trace_length : int, optional
            Maximum length of the trace before terminating the search. Defaults to 1000.
        """
        self.bprogram_generator = bprogram_generator
        self.max_trace_length = max_trace_length

    def verify(self):
        """
        Execute a DFS on the behavioral program to check for assertion errors.

        Returns:
        --------
        bool:
            True if no assertion error is encountered, otherwise False.
        Optional[tuple]:
            Trace leading up to the assertion error, if one is encountered. Otherwise, None.
        """
        dfs = DFSBProgram(self.bprogram_generator, max_trace_length=self.max_trace_length)
        try:
            _, _ = dfs.run()
            return True, None
        except BPAssertionError as e:
            return False, e.trace



from bppy.model.b_event import BEvent
from bppy.model.event_set import EventSet
from bppy.model.sync_statement import sync, choice
from bppy.utils.dfs import DFSBProgram
from bppy.utils.exceptions import BPAssertionError
from collections.abc import Iterable


class BProgramConverter:
	"""
	A class to convert a behavioral program to a PRISM model by exploring the DFS graph.
	"""
	def __init__(self, bprogram_generator, event_list, bt_names=None, max_trace_length=1000):
		"""
		Initialize a DFSBProgramVerifier.

		Parameters:
		-----------
		bprogram_generator : function
			A function that generates a new instance of the BProgram.
		event_list : list
			A list of event objects which are used in the model.
		bt_names : list, optional
			Names to used for the matching bthreads in the prism model.
			If none, the bthreads are named bt0, bt1, etc.
		max_trace_length : int, optional
			Maximum length of the trace before terminating the search. Defaults to 1000.
		"""
		self.bprogram_generator = bprogram_generator
		self.event_list = event_list
		if bt_names is not None:
			self.bt_names = bt_names
		else:
			names = [f'bt{i}' for i in range(len(self.bprogram_generator().bthreads))]
			self.bt_names = names
		self.max_trace_length = max_trace_length

	def collect_structure(self):
		"""
		Generates the tree structure and other necessary details of the behavioral program.
		"""
		dfs = DFSBProgram(self.bprogram_generator, self.event_list,
			max_trace_length=self.max_trace_length)
		init, mapper = dfs.run(explore_graph=False)

		names = [x.name for x in dfs.event_list]
		events = {e: BEvent(e) for e in names}
		bp_states = {self.bt_names[i]:v for i, v in mapper.items()}

		return names, events, bp_states

	def to_prism(self, output_file=None):
		"""
		Converts the behavioral program to a PRISM model of a markov decision process, with each bthread corresponding to a module.
		Returns the resulting text content.

		Parameters
		----------
		output_file : str, optional
			Name of the file to write the PRISM model to.
			Defaults to 'None'.
		"""

		event_names, events, bp_states = self.collect_structure()

		rule_template = "formula {}_{} = {};"
		bt_condition = "({}_{}_{}={})"

		header = 'mdp\n\n'

		req = [rule_template.format(e, "req",
				' | '.join([bt_condition.format(bt,'req',e,'true')
				for bt in self.bt_names]))
			for e in event_names]

		block = [rule_template.format(e, "block",
				' | '.join([bt_condition.format(bt,'block',e,'true')
				for bt in self.bt_names]))
			for e in event_names]

		enable = [rule_template.format(e, "enabled",
				' & '.join([f'({e}_req=true)',
							f'({e}_block=false)']))
			for e in event_names]

		labels = ["//event {} = {};".format(i, e)
			for i, e in enumerate(event_names)]
		
		event_transition = '[{}] ({}_enabled=true) -> 1: (event\'={});'
		guards = '\n\t'.join([event_transition.format(e, e, i)
						for i, e in enumerate(event_names)])
		module_main = f'\nmodule main\n\tevent: [-1..{len(event_names)}] init -1;'+\
						f'\n\t{guards}\nendmodule\n'

		header += '\n\n'.join(
			['\n'.join(sec) for sec in [req,block,enable,
				labels]])
		
		header += module_main
		header += '\n//-----------------------\n\n'

		def format_bt_module(name, event_names, bt_states):

			node_to_s = {s: i for i, s in enumerate(bp_states[name])}

			bt_req = {e: [] for e in event_names}
			bt_block = {e: [] for e in event_names}
			bt_trans = {}
			bt_probs = {}

			for node, n in node_to_s.items():
				if isinstance(node.data, sync):
					bt_trans[n] = {}
					for e in event_names:
						if ('request' in node.data):
							if (isinstance(node.data['request'], Iterable) and
									events[e] in node.data['request']):
								bt_req[e].append(n)
							elif (isinstance(node.data['request'], BEvent) and
									events[e] == node.data['request']):
								bt_req[e].append(n)
						if ('block' in node.data):
							if (isinstance(node.data['block'], Iterable) and
									events[e] in node.data['block']):
								bt_block[e].append(n)
							elif (isinstance(node.data['block'], BEvent) and
									events[e] == node.data['block']):
								bt_block[e].append(n)
						bt_trans[n][e] = (node_to_s[node.transitions[events[e]]] if
										events[e] in node.transitions else n)
				if isinstance(node.data, choice):
					bt_probs[n] = {choice_outcome: (node_to_s[node.transitions[choice_outcome][0]],
												choice_prob) for choice_outcome, choice_prob in node.data.options()}

			module_template = '\nmodule {}\n\t{}\n\n\t{}\nendmodule\n'
			state_template = "formula {}_{}_{} = {};"
			state_name = f's_{name}'
			state_init = f'{state_name}: [0..{len(bt_states)-1}] init 0;'
			event_transition = '[{}] ({}={}) -> 1: ({}\'={});'
			prob_transition = '[] ({}={}) -> {};'
			prob_format = '{}: ({}\'={})'

			req = [state_template.format(name, "req", e,
					' | '.join([('({}={})').format(state_name, n)
					for n in bt_req[e]]) if len(bt_req[e]) > 0 else 'false')
				for e in event_names]

			block = [state_template.format(name, "block", e,
					' | '.join([('({}={})').format(state_name, n)
					for n in bt_block[e]]) if len(bt_block[e]) > 0 else 'false')
				for e in event_names]

			transitions = []
			for n, tr in bt_trans.items():
				string_tr = [event_transition.format(e, state_name,
								n, state_name, s_tag) for e, s_tag in tr.items()]
				transitions.append('\n\t'.join(string_tr))
			
			probabilities = []
			for n, c in bt_probs.items():
				string_prob = [prob_format.format(p, state_name, next_state)
								for next_state, p in c.values()]
				probabilities.append(prob_transition.format(state_name, n,
									' + '.join(string_prob)))

			module = ''
			module += '\n\n'.join(
				['\n'.join(sec) for sec in [req,block]])

			module += module_template.format(name,
									state_init, '\n\t\n\t'.join(transitions + probabilities))

			return module


		content = '\n\n'.join([format_bt_module(bt_name, event_names,
								bp_states[bt_name]) for bt_name in self.bt_names])
		
		if output_file:
			with open(output_file, "w") as f:
				f.write(header + content)
		return header + content




from bppy.gym.bp_action_space import BPActionSpace
from bppy.gym.bp_env import BPEnv
from bppy.gym.bp_observation_space import BPObservationSpace
from bppy.gym.simple_bp_observation_space import SimpleBPObservationSpace



import unittest
import bppy as bp
from bppy.utils.dfs import DFSBThread, DFSBProgram
from bppy.utils.exceptions import BPAssertionError


class TestDFS(unittest.TestCase):

    def test_hello_world(self):
        @bp.thread
        def hello():
            yield bp.sync(request=bp.BEvent("Hello"))

        dfs = DFSBThread(lambda: hello(), bp.SimpleEventSelectionStrategy(), [bp.BEvent("Hello"), bp.BEvent("Dummy")])
        init_s, visited = dfs.run()
        assert len(visited) == 2

    def test_hot(self):
        @bp.thread
        def add_hot():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("HOT"))

        dfs = DFSBThread(lambda: add_hot(), bp.SimpleEventSelectionStrategy(),
                         [bp.BEvent("HOT"), bp.BEvent("COLD"), bp.BEvent("Dummy")])
        init_s, visited = dfs.run()
        assert len(visited) == 4

    def test_hot_rb(self):
        @bp.thread
        def add_hot():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("HOT"))

        dfs = DFSBThread(lambda: add_hot(), bp.SimpleEventSelectionStrategy(),
                         [bp.BEvent("HOT"), bp.BEvent("COLD"), bp.BEvent("Dummy")])
        init_s, visited, requested, blocked = dfs.run(return_requested_and_blocked=True)
        assert len(visited) == 4 and len(requested) == 1 and len(blocked) == 0

    def test_control(self):
        @bp.thread
        def control_temp():
            e = bp.BEvent("Dummy")
            while True:
                e = yield bp.sync(waitFor=bp.All(), block=e)

        dfs = DFSBThread(lambda: control_temp(), bp.SimpleEventSelectionStrategy(),
                         [bp.BEvent("HOT"), bp.BEvent("COLD")])
        init_s, visited = dfs.run()
        assert len(visited) == 3

    def test_bprogram(self):
        @bp.thread
        def add_hot():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("HOT"))

        @bp.thread
        def add_cold():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("COLD"))

        @bp.thread
        def control():
            while True:
                yield bp.sync(waitFor=bp.BEvent("HOT"))
                yield bp.sync(waitFor=bp.All(), block=bp.BEvent("HOT"))

        def bp_gen():
            return bp.BProgram(bthreads=[add_hot(), add_cold(), control()],
                               event_selection_strategy=bp.SimpleEventSelectionStrategy())

        dfs = DFSBProgram(bp_gen, [bp.BEvent("HOT"), bp.BEvent("COLD"), bp.BEvent("Dummy")])
        init_s, visited = dfs.run()
        assert len(visited) == 19

    def test_bprogram_no_list(self):
        @bp.thread
        def add_hot():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("HOT"))

        @bp.thread
        def add_cold():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("COLD"))

        @bp.thread
        def control():
            while True:
                yield bp.sync(waitFor=bp.BEvent("HOT"))
                yield bp.sync(waitFor=bp.All(), block=bp.BEvent("HOT"))

        def bp_gen():
            return bp.BProgram(bthreads=[add_hot(), add_cold(), control()],
                               event_selection_strategy=bp.SimpleEventSelectionStrategy())

        dfs = DFSBProgram(bp_gen)
        init_s, visited = dfs.run()
        assert len(visited) == 19

    def test_bp_assertion_error(self):
        @bp.thread
        def add_hot():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("HOT"))

        @bp.thread
        def add_cold():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("COLD"))
            assert False

        @bp.thread
        def control():
            while True:
                yield bp.sync(waitFor=bp.BEvent("HOT"))
                yield bp.sync(waitFor=bp.All(), block=bp.BEvent("HOT"))

        def bp_gen():
            return bp.BProgram(bthreads=[add_hot(), add_cold(), control()],
                               event_selection_strategy=bp.SimpleEventSelectionStrategy())

        dfs = DFSBProgram(bp_gen)
        try:
            init_s, visited = dfs.run()
        except BPAssertionError:
            pass
        assert True



from copy import copy
from bppy.model.sync_statement import sync, choice
from warnings import warn

def execution_thread(func):
    return thread(func, 'execution')

def analysis_thread(func):
    return thread(func, 'analysis')

def thread(func, mode='execution'):
    """
    A decorator to wrap bthread generator with, in order to handle data transmission and bthread termination.
    """
    def wrapper(*args):
        while True:
            m = None
            f = func(*args)
            while True:
                try:
                    e = f.send(m)
                    if type(e) == dict:
                        warn("using dict for statements is deprecated, use bppy.model.sync_statement.sync instead.")
                        e = sync(**e)
                    if isinstance(e, sync):
                        local_vars = {var:val for var, val in copy(f.gi_frame.f_locals).items()}
                        e["locals"] = copy(local_vars)
                        m = yield e
                        if m is None:
                            break
                    elif isinstance(e, choice):
                        if mode == 'execution':
                            sample = e.sample()
                            m = sample
                        else:
                            m = yield e
                    else:
                        raise TypeError("bthread must yield a bppy.model.sync_statement object")
                except (KeyError, StopIteration):
                    m = yield None
                    break
    return wrapper


def b_thread(func):
    warn("the b_thread decorator is deprecated, use thread instead.")
    return thread(func)





from bppy.model.event_selection.smt_event_selection_strategy import SMTEventSelectionStrategy
from bppy.utils.z3helper import *


class RichEventSelectionStrategy(SMTEventSelectionStrategy):
    """
    A solver-based implementation of EventSelectionStrategy using z3 SMT solver. The strategy used is the one
    presented in the paper `Executing Scenario-Based Specification with Dynamic Generation of Rich Events
    <https://www.wisdom.weizmann.ac.il/~dharel/papers/CCIS2019RichEvents.pdf>`_.
    """

    def is_satisfied(self, event, statement):
        return is_true(event.eval(statement.get('waitFor', true)))




import unittest
import bppy as bp
from bppy.gym import *
import numpy as np


class TestGym(unittest.TestCase):

    def test_env(self):
        @bp.thread
        def add_hot():  # request hot 5 times, and specify a reward
            for i in range(5):
                yield bp.sync(request=bp.BEvent("HOT"), localReward=-0.01)
            yield bp.sync(waitFor=bp.All(), localReward=1)

        @bp.thread
        def add_cold():  # request cold 5 times
            for i in range(5):
                yield bp.sync(request=bp.BEvent("COLD"))

        @bp.thread
        def control():  # blocks HOT from occurring twice in a row
            while True:
                yield bp.sync(waitFor=bp.BEvent("HOT"))
                yield bp.sync(block=bp.BEvent("HOT"), waitFor=bp.BEvent("COLD"))

        def init_bprogram():
            return bp.BProgram(bthreads=[add_hot(), add_cold(), control()],
                            event_selection_strategy=bp.SimpleEventSelectionStrategy())

        env = BPEnv(bprogram_generator=init_bprogram,
                    action_list=[bp.BEvent("HOT"), bp.BEvent("COLD"), bp.BEvent("DONE")])

        state, _ = env.reset()
        assert (state == np.asarray([0, 0, 0])).all()
        state, reward, terminated, _, info = env.step(0)
        assert (state == np.asarray([1, 0, 1])).all()



from gymnasium.spaces import MultiDiscrete


class BPObservationSpace(MultiDiscrete):
    """
    A base class used to represent a BProgram-based observation space. This is an abstract class
    that requires the implementation of `bp_state_to_gym_space` methods.
    """
    @property
    def np_random(self):
        return super().np_random

    @property
    def shape(self):
        return super().shape

    def sample(self):
        pass

    def seed(self, seed=None):
        return super().seed(seed)

    def contains(self, x):
        pass

    def __contains__(self, x):
        return super().__contains__(x)

    def __setstate__(self, state):
        super().__setstate__(state)

    def to_jsonable(self, sample_n):
        return super().to_jsonable(sample_n)

    def from_jsonable(self, sample_n):
        return super().from_jsonable(sample_n)

    def bp_state_to_gym_space(self, bthreads_states):
        """
        Abstract method that transforms the bprogram's state, received as a list of bthreads statements, to a gym space
        representation.
        """
        raise NotImplementedError("bp_state_to_gym_space not implemented")


from bppy.gym.bp_observation_space import BPObservationSpace
import numpy as np
from bppy.utils.dfs import DFSBThread


class SimpleBPObservationSpace(BPObservationSpace):
    """
    Defines a simple observation space for BPrograms that inherit from
    :class:`BPObservationSpace <bppy.gym.bp_observation_space.BPObservationSpace>`. This class is responsible for
    mapping the state of the BProgram's bthreads to a Gym-compatible observation space.

    Attributes
    ----------
    bprogram_generator : callable
        A function that generates an instance of the BProgram.
    event_list : list
        A list of possible events in the BProgram.

    """
    def __init__(self, bprogram_generator, event_list, dtype=np.int64, seed=None):
        """
        Initializes the :class:`SimpleBPObservationSpace <bppy.gym.simaple_bp_observation_space.SimpleBPObservationSpace>`.

        Parameters
        ----------
        bprogram_generator : callable
            A function that generates an instance of the BProgram.
        event_list : list
            A list of possible events in the BProgram.
        dtype : data-type, optional
            Data type for the observation space, defaults to `np.int64`.
        seed : `int` or `None`, optional
            Seed for the random number generator, defaults to None.
        """
        self.bprogram_generator = bprogram_generator
        self.event_list = event_list
        self.mapper = self.compute_bthread_state_space()
        self.bt_state_sizes = np.asarray([len(self.mapper[x]) for x in self.mapper], dtype=dtype)
        super(SimpleBPObservationSpace, self).__init__(self.bt_state_sizes, dtype, seed)

    def bp_state_to_gym_space(self, bthreads_states):
        """
        Converts the state of the BProgram's bthreads to the Gym-compatible observation space.

        Parameters
        ----------
        bthreads_states : list
            A list representing the bthreads current statements.

        Returns
        -------
        gym_space : `np.ndarray`
            An array representing the Gym-compatible observation space.
        """
        return np.asarray([x.index(y) for x, y in zip(self.mapper.values(), bthreads_states)], dtype=self.dtype)

    def compute_bthread_state_space(self):
        mapper = {}
        n = len(self.bprogram_generator().bthreads)
        ess = self.bprogram_generator().event_selection_strategy
        for i in range(n):
            f = lambda: self.bprogram_generator().bthreads[i]
            dfs = DFSBThread(f, ess, self.event_list)
            init_s, visited = dfs.run()
            visited = [s.data for s in visited]
            mapper[i] = visited
        return mapper


import unittest
import bppy as bp
from bppy.utils.z3helper import *


class TestExamples(unittest.TestCase):

    def test_hello_world(self):
        @bp.thread
        def hello():  # requests "Hello" once
            yield bp.sync(request=bp.BEvent("Hello"))

        @bp.thread
        def world():  # requests "World" once
            yield bp.sync(request=bp.BEvent("World"))

        b_program = bp.BProgram(bthreads=[hello(), world()],
                                event_selection_strategy=bp.SimpleEventSelectionStrategy())
        b_program.run()
        assert True

    def test_hot_cold_all(self):
        @bp.thread
        def add_hot():  # requests "HOT" three times
            yield bp.sync(request=bp.BEvent("HOT"))
            yield bp.sync(request=bp.BEvent("HOT"))
            yield bp.sync(request=bp.BEvent("HOT"))

        @bp.thread
        def add_cold():  # requests "COLD" three times
            yield bp.sync(request=bp.BEvent("COLD"))
            yield bp.sync(request=bp.BEvent("COLD"))
            yield bp.sync(request=bp.BEvent("COLD"))

        @bp.thread
        def control_temp():
            # This b-thread controls the temperature by blocking the previously selected event
            # and waiting for all other events in each iteration of its loop
            e = bp.BEvent("Dummy")
            while True:
                e = yield bp.sync(waitFor=bp.All(), block=e)

        b_program = bp.BProgram(bthreads=[add_hot(), add_cold(), control_temp()],
                                event_selection_strategy=bp.SimpleEventSelectionStrategy())
        b_program.run()
        assert True

    def test_hot_cold_bath(self):
        @bp.thread
        def add_hot():  # requests "HOT" three times
            yield bp.sync(request=bp.BEvent("HOT"))
            yield bp.sync(request=bp.BEvent("HOT"))
            yield bp.sync(request=bp.BEvent("HOT"))

        @bp.thread
        def add_cold():  # requests "COLD" three times
            yield bp.sync(request=bp.BEvent("COLD"))
            yield bp.sync(request=bp.BEvent("COLD"))
            yield bp.sync(request=bp.BEvent("COLD"))

        @bp.thread
        def control_temp():
            while True:
                yield bp.sync(waitFor=bp.BEvent("COLD"), block=bp.BEvent("HOT"))
                yield bp.sync(waitFor=bp.BEvent("HOT"), block=bp.BEvent("COLD"))

        b_program = bp.BProgram(bthreads=[add_hot(), add_cold(), control_temp()],
                                event_selection_strategy=bp.SimpleEventSelectionStrategy())
        b_program.run()
        assert True

    def test_hot_cold_dynamic(self):
        @bp.thread
        def add_hot():  # requests "HOT" three times
            yield bp.sync(request=bp.BEvent("HOT"))
            yield bp.sync(request=bp.BEvent("HOT"))
            yield bp.sync(request=bp.BEvent("HOT"))

        @bp.thread
        def add_cold():  # requests "COLD" three times
            yield bp.sync(request=bp.BEvent("COLD"))
            yield bp.sync(request=bp.BEvent("COLD"))
            yield bp.sync(request=bp.BEvent("COLD"))

        @bp.thread
        def control_temp(block_event):
            # This b-thread blocks the event given as an argument and then adds
            # a new b-thread with the previously selected event
            block_event = yield bp.sync(block=block_event, waitFor=bp.All())
            b_program.add_bthread(control_temp(block_event))

        b_program = bp.BProgram(bthreads=[add_hot(), add_cold(), control_temp(bp.BEvent("HOT"))],
                                event_selection_strategy=bp.SimpleEventSelectionStrategy())
        b_program.run()
        assert True

    def test_hot_cold_smt(self):
        # Defining SMT variables
        hot = Bool('hot')
        cold = Bool('cold')
        temp = Real('temp')

        @bp.thread
        def three_hot():  # requesting that the hot variable will be true three times
            for i in range(3):
                yield bp.sync(request=hot)

        @bp.thread
        def three_cold():  # requesting that the cold variable will be true three times
            for j in range(3):
                yield bp.sync(request=cold)

        @bp.thread
        def control_temp():  # blocking consecutive assignments where hot is true
            while True:
                yield bp.sync(waitFor=hot)
                yield bp.sync(block=hot, waitFor=Not(hot))

        @bp.thread
        def mutual_exclusion():  # A b-thread that ensures that cold and hot cannot be both true at the same time
            yield bp.sync(block=And(cold, hot), waitFor=false)

        @bp.thread
        def hot_temp():  # A b-thread that ensures that temp is above 50 when hot is true
            yield bp.sync(block=And(hot, temp <= 50), waitFor=false)

        @bp.thread
        def cold_temp():  # A b-thread that ensures that temp is below 50 when cold is true
            yield bp.sync(block=And(cold, temp >= 50), waitFor=false)

        @bp.thread
        def after_hot_temp():  # A b-thread that ensures that temp is above 20 when the previous event was hot
            while True:
                yield bp.sync(waitFor=hot)
                yield bp.sync(block=temp <= 20, waitFor=Not(hot))

        @bp.thread
        def after_cold_temp():  # A b-thread that ensures that temp is below 80 when the previous event was cold
            while True:
                yield bp.sync(waitFor=cold)
                yield bp.sync(block=temp >= 80, waitFor=Not(cold))

        b_program = bp.BProgram(bthreads=[three_cold(),
                                          three_hot(),
                                          mutual_exclusion(),
                                          control_temp(),
                                          hot_temp(),
                                          cold_temp(),
                                          after_hot_temp(),
                                          after_cold_temp()],
                                event_selection_strategy=bp.SMTEventSelectionStrategy())
        b_program.run()
        assert True

    def test_external_events(self):
        class External(bp.BEvent):
            pass

        any_external = bp.EventSet(lambda event: isinstance(event, External))

        @bp.thread
        def add_external():
            b_program.enqueue_external_event(External("A"))
            b_program.enqueue_external_event(External("B"))
            b_program.enqueue_external_event(External("C"))
            while True:
                yield bp.sync(waitFor=bp.All())

        @bp.thread
        def act_on_external():
            while True:
                # triggers external events if exists, else terminates the bprogram
                event = yield bp.sync(block=bp.All(), waitFor=any_external)
                yield bp.sync(request=bp.BEvent(event.name))

        b_program = bp.BProgram(bthreads=[add_external(), act_on_external()],
                                event_selection_strategy=bp.SimpleEventSelectionStrategy())
        b_program.run()
        assert True

    def test_tic_tac_toe_priorities(self):
        # BEvents functions for x and o moves
        x = lambda row, col: bp.BEvent('X' + str(row) + str(col))
        o = lambda row, col: bp.BEvent('O' + str(row) + str(col))

        # locations of all possible lines
        LINES = [[(i, j) for j in range(3)] for i in range(3)] + [[(i, j) for i in range(3)] for j in range(3)] + [
            [(i, i) for i in range(3)]] + [[(i, 3 - i - 1) for i in range(3)]]
        x_lines = [[x(i, j) for (i, j) in line] for line in LINES]
        o_lines = [[o(i, j) for (i, j) in line] for line in LINES]

        # define events sets of x an o moves
        any_x = [x(i, j) for i in range(3) for j in range(3)]
        any_o = [o(i, j) for i in range(3) for j in range(3)]
        move_events = bp.EventSet(lambda e: e.name.startswith('X') or e.name.startswith('O'))

        # define static terminal events
        static_event = {
            'OWin': bp.BEvent('OWin'),
            'XWin': bp.BEvent('XWin'),
            'draw': bp.BEvent('Draw')
        }

        @bp.thread
        def square_taken(row, col):  # blocks moves to a square that is already taken
            yield bp.sync(waitFor=[x(row, col), o(row, col)])
            yield bp.sync(block=[x(row, col), o(row, col)])

        @bp.thread
        def enforce_turns():  # blocks moves that are not in turn
            while True:
                yield bp.sync(waitFor=any_x, block=any_o)
                yield bp.sync(waitFor=any_o, block=any_x)

        @bp.thread
        def end_of_game():  # blocks moves after the game is over
            yield bp.sync(waitFor=list(static_event.values()))
            yield bp.sync(block=bp.All())

        @bp.thread
        def detect_draw():  # detects a draw
            for r in range(3):
                for c in range(3):
                    yield bp.sync(waitFor=move_events)
            yield bp.sync(request=static_event['draw'], priority=90)

        @bp.thread
        def detect_x_win(line):  # detects a win by player X
            for i in range(3):
                yield bp.sync(waitFor=line)
            yield bp.sync(request=static_event['XWin'], priority=100)

        @bp.thread
        def detect_o_win(line):  # detects a win by player O
            for i in range(3):
                yield bp.sync(waitFor=line)
            yield bp.sync(request=static_event['OWin'], priority=100)

        @bp.thread
        def center_preference():
            while True:
                yield bp.sync(request=o(1, 1), priority=35)

        @bp.thread
        def corner_preference():
            while True:
                yield bp.sync(request=[o(0, 0), o(0, 2), o(2, 0), o(2, 2)], priority=20)

        @bp.thread
        def side_preference():
            while True:
                yield bp.sync(request=[o(0, 1), o(1, 0), o(1, 2), o(2, 1)], priority=10)

        @bp.thread
        def add_third_o(line):
            for i in range(2):
                yield bp.sync(waitFor=line)
            yield bp.sync(request=line, priority=50)

        @bp.thread
        def prevent_third_x(xline, oline):
            for i in range(2):
                yield bp.sync(waitFor=xline)
            yield bp.sync(request=oline, priority=40)

        @bp.thread
        def block_fork(xfork, ofork):  # blocks a fork strategy
            for i in range(2):
                yield bp.sync(waitFor=xfork)
            yield bp.sync(request=ofork, priority=30)

        forks22 = [[x(1, 2), x(2, 0)], [x(2, 1), x(0, 2)], [x(1, 2), x(2, 1)]], [o(2, 2), o(0, 2), o(2, 0)]
        forks02 = [[x(1, 2), x(0, 0)], [x(0, 1), x(2, 2)], [x(1, 2), x(0, 1)]], [o(0, 2), o(0, 0), o(2, 2)]
        forks20 = [[x(1, 0), x(2, 2)], [x(2, 1), x(0, 0)], [x(2, 1), x(1, 0)]], [o(2, 0), o(0, 0), o(2, 2)]
        forks00 = [[x(0, 1), x(2, 0)], [x(1, 0), x(0, 2)], [x(0, 1), x(1, 0)]], [o(0, 0), o(0, 2), o(2, 0)]
        forks_diag = [[x(0, 2), x(2, 0)], [x(0, 0), x(2, 2)]], [o(0, 1), o(1, 0), o(1, 2), o(2, 1)]

        @bp.thread
        def player_x():  # simulate player X
            while True:
                yield bp.sync(request=any_x)

        bprog = bp.BProgram(
            bthreads=[square_taken(i, j) for i in range(3) for j in range(3)] +
                     [enforce_turns(), end_of_game(), detect_draw()] +
                     [detect_x_win(line) for line in x_lines] +
                     [detect_o_win(line) for line in o_lines] +
                     [center_preference(), corner_preference(), side_preference()] +
                     [add_third_o(line) for line in o_lines] +
                     [prevent_third_x(xline, oline) for (xline, oline) in zip(x_lines, o_lines)] +
                     [block_fork(xfork, forks22[1]) for xfork in forks22[0]] +
                     [block_fork(xfork, forks02[1]) for xfork in forks02[0]] +
                     [block_fork(xfork, forks20[1]) for xfork in forks20[0]] +
                     [block_fork(xfork, forks00[1]) for xfork in forks00[0]] +
                     [block_fork(xfork, forks_diag[1]) for xfork in forks_diag[0]] +
                     [player_x()],
            event_selection_strategy=bp.PriorityBasedEventSelectionStrategy(default_priority=0))
        bprog.run()
        assert True



from bppy.model.b_event import BEvent
from bppy.model.sync_statement import sync, choice
from bppy.utils.exceptions import BPAssertionError

class Node:
    def __init__(self, prefix, data):
        self.prefix = prefix
        self.data = data
        self.transitions = {}
        self.__hash = None

    def __key(self):
        if isinstance(self.data, choice):
            return self.data._id
        if not self.__hash:
            self.__hash = hash(str(self.data))
            return self.__hash
        else:
            return self.__hash
        return str(self.data)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if self.__key() == other.__key():
            if not (isinstance(self.data, choice) or isinstance(other.data, choice)):
                return self.data == other.data
            else:
                return True
        else:
            return False

    def __str__(self):
        return str(self.prefix) + str(self.data)

    def __repr__(self):
        return self.__str__()

    def get_key(self):
        return self.__key()


class DFSBThread:
    def __init__(self, bthread_gen, ess, event_list):
        self.bthread_gen = bthread_gen
        self.ess = ess
        self.event_list = event_list

    def get_state(self, prefix):
        bt = self.bthread_gen()
        s = bt.send(None) # s = sync or Choice
        for e in prefix: # e = event or choice sample result
            if s is None:
                break
            if isinstance(s, sync): 
                if 'block' in s:
                    if isinstance(s.get('block'), BEvent):
                        if e == s.get('block'):
                            return None
                    else:
                        if e in s.get('block'):
                            return None
                if self.ess.is_satisfied(e, s):
                    s = bt.send(e)
            elif isinstance(s, choice):
                s = bt.send(e)
        if s is None:
            return sync()
        return s

    def run(self, return_requested_and_blocked=False):
        init_s = Node(tuple(), self.get_state(tuple()))
        visited = []
        stack = []
        stack.append(init_s)
        requested = set()
        blocked = set()

        while len(stack):
            s = stack.pop()
            if s not in visited:
                visited.append(s)
            if isinstance(s.data, sync) and return_requested_and_blocked:
                if "request" in s.data:
                    if isinstance(s.data["request"], BEvent):
                        requested.add(s.data["request"])
                    else:
                        requested.update(s.data["request"])
                if "block" in s.data:
                    if isinstance(s.data["block"], BEvent):
                        blocked.add(s.data["block"])
                    else:
                        blocked.update([x for x in self.event_list if x in s.data["block"]])
            if isinstance(s.data, choice):
                for outcome, prob in s.data.options():
                    if isinstance(outcome, dict) and len(outcome) == 1:
                        outcome = outcome[0]
                    new_s = Node(s.prefix + (outcome,), 
                        self.get_state(s.prefix + (outcome,)))
                    s.transitions[outcome] = (new_s, prob)
                    if new_s not in visited:
                        stack.append(new_s)
            if isinstance(s.data, sync):
                for e in self.event_list:
                    if self.ess.is_satisfied(e, s.data):
                        new_s = Node(s.prefix + (e,), self.get_state(s.prefix + (e,)))
                        if new_s not in visited:
                            stack.append(new_s)
                    else:
                        new_s = s
                    if new_s.data is None:
                        continue
                    s.transitions[e] = new_s
        if return_requested_and_blocked:
            return init_s, visited, requested, blocked
        return init_s, visited


class NodeList:
    def __init__(self, nodes, prefix):
        self.nodes = nodes
        self.prefix = prefix
        self.transitions = {}

    def __key(self):
        return hash(tuple([n.get_key() for n in self.nodes]))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return self.__key() == other.__key()


class DFSBProgram:
    def __init__(self, bprogram_generator, event_list=None, max_trace_length=1000):
        self.bprogram_generator = bprogram_generator
        self.event_list = event_list
        self.max_trace_length = max_trace_length

    def run(self, explore_graph=True):
        if self.event_list:
            mapper = {}
            init = []
            n = len(self.bprogram_generator().bthreads)
            ess = self.bprogram_generator().event_selection_strategy
            for i in range(n):
                f = lambda: self.bprogram_generator().bthreads[i]
                dfs = DFSBThread(f, ess, self.event_list)
                init_s, visited = dfs.run()
                mapper[i] = visited
                init.append(init_s)

            if not explore_graph: # stop before mapping sync statements
                return init, mapper 

            init = NodeList(init, tuple())
            visited = set()
            stack = [init]
            while len(stack):
                s = stack.pop()
                if s not in visited:
                    visited.add(s)

                for e in ess.selectable_events([x.data for x in s.nodes if x.data is not None]):
                    new_s = []
                    for i, bt_s in enumerate(s.nodes):
                        s_temp = mapper[i][mapper[i].index(bt_s)]
                        new_s.append(s_temp.transitions[e])
                    new_s = NodeList(new_s, s.prefix + (e, ))
                    s.transitions[e] = new_s
                    if new_s not in visited:
                        stack.append(new_s)
            visited = list(visited)
            visited.remove(init)
            visited.insert(0, init)
            return init, visited
        else:
            ess = self.bprogram_generator().event_selection_strategy
            bprogram = self.bprogram_generator()
            bprogram.setup()
            init = NodeList([Node(tuple(), t) for t in self.tickets_without_bt(bprogram.tickets)], tuple())
            visited = set()
            stack = [init]
            while len(stack):
                s = stack.pop()
                if s not in visited:
                    visited.add(s)

                for e in ess.selectable_events([x.data for x in s.nodes if x.data is not None]):
                    bprogram = self.bprogram_generator()
                    bprogram.setup()
                    for pre_e in s.prefix:
                        bprogram.advance_bthreads(bprogram.tickets, pre_e)
                    try:
                        bprogram.advance_bthreads(bprogram.tickets, e)
                    except AssertionError:
                        raise BPAssertionError("Assertion error in DFSBProgram", s.prefix + (e,))
                    new_s = NodeList([Node(s.prefix + (e,), t) for t in self.tickets_without_bt(bprogram.tickets)], s.prefix + (e,))
                    s.transitions[e] = new_s
                    if (new_s not in visited) and (len(new_s.prefix) <= self.max_trace_length):
                        stack.append(new_s)
            visited = list(visited)
            visited.remove(init)
            visited.insert(0, init)
            return init, visited

    @staticmethod
    def tickets_without_bt(tickets):
        return [dict([(k, v) for k, v in t.items() if k != 'bt']) for t in tickets]






import unittest
import bppy as bp


class TestBProgram(unittest.TestCase):

    def test_advance_bthreads(self):
        @bp.thread
        def hello():  # requests "Hello" once
            yield bp.sync(request=bp.BEvent("Hello"))

        @bp.thread
        def world():  # requests "World" once
            yield bp.sync(request=bp.BEvent("World"))

        b_program = bp.BProgram(bthreads=[hello(), world()],
                             event_selection_strategy=bp.SimpleEventSelectionStrategy())

        b_program.setup()
        b_program.advance_bthreads(b_program.tickets, bp.BEvent("Hello"))
        b_program.advance_bthreads(b_program.tickets, bp.BEvent("World"))
        assert all([len(x) == 0 for x in b_program.tickets]) and len(b_program.tickets) == 2



from random import choices
import itertools

# from https://stackoverflow.com/a/43649323

def weighted_sample_without_replacement(population, weights, k=1):
    weights = list(weights)
    positions = range(len(population))
    indices = []
    while True:
        needed = k - len(indices)
        if not needed:
            break
        for i in choices(positions, weights, k=needed):
            if weights[i]:
                weights[i] = 0.0
                indices.append(i)
    return [population[i] for i in indices]

def sequence_probability_nr_s(choice, sequence):
    total_prob = 0
    for perm in itertools.permutations(sequence):
        left_prob = 1
        cum_prob = 1
        for event in perm:
            cum_prob *= (choice[event])/left_prob
            left_prob -= choice[event]
        total_prob += cum_prob
    return (sequence, total_prob)


from gymnasium.spaces import Discrete
import numpy as np


class BPActionSpace(Discrete):
    """
    Defines a custom action space for BPrograms, which inherits from the Discrete space class. The actions correspond to
    selectable events in the BProgram, and this class helps to manage and validate those actions within the environment.

    Attributes
    ----------
    event_list : list
        A list of possible events in the BProgram.

    """
    def __init__(self, event_list):
        """
        Initializes the BPActionSpace.

        Parameters
        ----------
        event_list : list
            A list of possible events in the BProgram.
        """
        self.event_list = event_list
        self.bprogram = None
        Discrete.__init__(self, len(event_list))

    def sample(self):
        """
        Randomly samples an action from the set of possible actions.

        Returns
        -------
        action : int
            The index of the randomly selected action.
        """
        return self.np_random.choice(self._possible_actions())

    def contains(self, x):
        """
        Checks whether the provided value corresponds to a valid action.

        Parameters
        ----------
        x : int, np.generic, or np.ndarray
            The value to check for validity.

        Returns
        -------
        valid : bool
            True if the value is a valid action, False otherwise.
        """
        if isinstance(x, int):
            as_int = x
        elif isinstance(x, (np.generic, np.ndarray)) and (x.dtype.char in np.typecodes['AllInteger'] and x.shape == ()):
            as_int = int(x)
        else:
            return False
        return as_int in self._possible_actions()

    def __repr__(self):
        return "BPActionSpace(%d)" % self.n

    def __eq__(self, other):
        return isinstance(other, BPActionSpace) and self.bprogram == other.bprogram and self.event_list == other.event_list

    def _possible_actions(self):
        possible_events = self.bprogram.event_selection_strategy.selectable_events(self.bprogram.tickets)
        return [k for k, v in enumerate(self.event_list) if v in possible_events]



import unittest
import bppy as bp
from bppy.analysis.bprogram_converter import BProgramConverter

class TestBProgram(unittest.TestCase):

	def test_prism_conversion(self):
		@bp.thread
		def main():
			r = yield bp.choice({'a': 0.2, 'b': 0.3, 'c': 0.5})
			yield bp.sync(request=bp.BEvent(f'event_{r}'))

		def bp_gen():
			return bp.BProgram(bthreads=[main()],
							event_selection_strategy=bp.SimpleEventSelectionStrategy())

		converter = BProgramConverter(bp_gen, [bp.BEvent('a'), bp.BEvent('b'), bp.BEvent('c')])
		output = converter.to_prism(None)
		assert output



import unittest
import bppy as bp
from bppy.analysis.dfs_bprogram_verifier import DFSBProgramVerifier
from bppy.analysis.symbolic_bprogram_verifier import SymbolicBProgramVerifier


class TestAnalysis(unittest.TestCase):

    def test_dfs_failed(self):
        @bp.thread
        def add_hot():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("HOT"))

        @bp.thread
        def add_cold():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("COLD"))

        @bp.thread
        def control():
            while True:
                yield bp.sync(waitFor=bp.BEvent("HOT"))
                yield bp.sync(waitFor=bp.All(), block=bp.BEvent("HOT"))

        @bp.thread
        def check():
            yield bp.sync(waitFor=bp.BEvent("HOT"))
            yield bp.sync(waitFor=bp.BEvent("HOT"))
            yield bp.sync(waitFor=bp.BEvent("HOT"))
            assert False

        def bp_gen():
            return bp.BProgram(bthreads=[add_hot(), add_cold(), control(), check()],
                               event_selection_strategy=bp.SimpleEventSelectionStrategy())

        ver = DFSBProgramVerifier(bp_gen)
        ok, counter_example = ver.verify()
        assert not ok

    def test_dfs_passed(self):
        @bp.thread
        def add_hot():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("HOT"))

        @bp.thread
        def add_cold():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("COLD"))

        @bp.thread
        def control():
            while True:
                yield bp.sync(waitFor=bp.BEvent("HOT"))
                yield bp.sync(waitFor=bp.All(), block=bp.BEvent("HOT"))

        @bp.thread
        def check():
            yield bp.sync(waitFor=bp.BEvent("HOT"))
            yield bp.sync(waitFor=bp.BEvent("HOT"))
            yield bp.sync(waitFor=bp.BEvent("HOT"))
            yield bp.sync(waitFor=bp.BEvent("HOT"))
            assert False

        def bp_gen():
            return bp.BProgram(bthreads=[add_hot(), add_cold(), control(), check()],
                               event_selection_strategy=bp.SimpleEventSelectionStrategy())

        ver = DFSBProgramVerifier(bp_gen)
        ok, counter_example = ver.verify()
        assert ok

    def test_symbolic_pass(self):

        @bp.thread
        def add_hot():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("HOT"))
            yield bp.sync(request=bp.BEvent("FINISH"), block={bp.BEvent("COLD")})

        @bp.thread
        def add_cold():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("COLD"))

        @bp.thread
        def control():
            while True:
                yield bp.sync(waitFor=bp.BEvent("HOT"), block=bp.EventSet(lambda e: e.name == "COLD"))
                yield bp.sync(waitFor=bp.All(), block=bp.BEvent("HOT"))

        def bp_gen():
            return bp.BProgram(bthreads=[add_hot(), add_cold(), control()])

        verifier = SymbolicBProgramVerifier(bp_gen, [bp.BEvent("HOT"), bp.BEvent("COLD"), bp.BEvent("FINISH")])
        result, explanation_str = verifier.verify(spec="F event = FINISH", type="BDD", bound=1000,
                                                  find_counterexample=True, print_info=False)

        assert result

    def test_symbolic_fail(self):

        @bp.thread
        def add_hot():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("HOT"))
            yield bp.sync(request=bp.BEvent("FINISH"), block=[bp.BEvent("COLD")])

        @bp.thread
        def add_cold():
            for i in range(3):
                yield bp.sync(request=bp.BEvent("COLD"))

        @bp.thread
        def control():
            while True:
                yield bp.sync(waitFor=bp.BEvent("HOT"))
                yield bp.sync(waitFor=bp.All(), block=bp.BEvent("HOT"))

        def bp_gen():
            return bp.BProgram(bthreads=[add_hot(), add_cold(), control()])

        verifier = SymbolicBProgramVerifier(bp_gen, [bp.BEvent("HOT"), bp.BEvent("COLD"), bp.BEvent("FINISH")])
        result, explanation_str = verifier.verify(spec="F event = FINISH", type="BMC", bound=100,
                                                  find_counterexample=True, print_info=False)

        assert not result and explanation_str is not None



import setuptools
import os

def get_version(rel_path):
    with open(rel_path, "r") as fh:
        for line in fh:
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
with open(os.path.join(__location__, "README.md"), "r") as fh:
    long_description = fh.read()

gym_dependencies = [
    "gymnasium>=0.28.1",
    "stable-baselines3>=2.0.0"
    ]
develop_dependencies = [
    "Sphinx>=5.3.0",
    "sphinx-rtd-theme>=1.2.1",
    "graphviz>=0.20.1",
    "pygame>=2.1.2"
]

setuptools.setup(
    name="bppy",
    version=get_version("bppy/__init__.py"),
    author="Tom Yaacov",
    author_email="tomyaacov1210@gmail.com",
    description="BPpy: Behavioral Programming In Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bThink-BGU/BPpy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "z3-solver>=4.8.5.0"
    ],
    extras_require={
        "gym": gym_dependencies,
        "develop": gym_dependencies + develop_dependencies
    },
    python_requires='>=3.4'
)


from bppy.utils.dfs import DFSBThread
from bppy.model.sync_statement import request, block, waitFor
from bppy.model.b_event import BEvent
from bppy.model.event_selection.simple_event_selection_strategy import SimpleEventSelectionStrategy
import sys
import tempfile
try:
    import pynusmv
    from pynusmv.model import *
    from pynusmv.mc import check_ltl_spec, check_explain_ltl_spec
    from pynusmv.bmc.glob import bmc_setup, BmcSupport, master_be_fsm
    from pynusmv.parser import parse_ltl_spec
    from pynusmv.node import Node
    from pynusmv.sat import SatSolverFactory, Polarity, SatSolverResult
    from pynusmv.bmc import ltlspec, utils as bmcutils
except ImportError:
    raise ImportError("PyNuSMV is not installed. Please install it to use SymbolicBProgramVerifier. More info on "
                      "PyNuSMV installation can be found at https://github.com/LouvainVerificationLab/pynusmv")


class SymbolicBProgramVerifier:
    """
    A class to verify a behavioral program symbolically using `PyNuSMV`. The verifier can operate in two modes:
    Binary Decision Diagrams (BDD) and SAT-based Bounded Model Checking (BMC).

    *Note:*

    1. The class requires the installation of `PyNuSMV`. More info on its installation can be found at `https://github.com/LouvainVerificationLab/pynusmv
    <https://github.com/LouvainVerificationLab/pynusmv>`_.

    2. The verifier is currently limited to b-programs with :class:`SimpleEventSelectionStrategy
    <bppy.model.event_selection.event_selection_strategy.SimpleEventSelectionStrategy>`.

    3. The verifier does not support events with data.

    Attributes:
    -----------
    bprogram_generator : function
        A function that generates a new instance of the BProgram.
    event_list : list
        List of all possible events in the system.
    """
    def __init__(self, bprogram_generator, event_list):
        self.bprogram_generator = bprogram_generator
        self.event_list = event_list
        self.bthread_num = len(self.bprogram_generator().bthreads)
        sys.setrecursionlimit(10000)

    def verify(self, spec, type="BDD", bound=1000, find_counterexample=False, print_info=False):
        """
        Verifies a given specification against the behavioral program.

        Parameters:
        -----------
        spec : str
            The specification to be verified, in `NuSMV <https://nusmv.fbk.eu/>`_ LTL specification format.
        type : str, optional
            The verification type to be used: Binary Decision Diagrams ("BDD") or SAT-based Bounded Model Checking ("BMC").
            Default is "BDD".
        bound : int, optional
            For BMC, the maximum number of steps to be considered. Default is 1000.
        find_counterexample : bool, optional
            If True and a specification violation is found, a counterexample is produced.
            Default is False.
        print_info : bool, optional
            If True, information about the verification process is printed. Default is False.

        Returns:
        --------
        bool:
            True if the specification is satisfied, False otherwise.
        Optional[str]:
            A counterexample, if one exists and find_counterexample is True. Otherwise, None.
        """
        if print_info:
            print("initializing NuSMV")
        pynusmv.init.init_nusmv()
        if print_info:
            print("Converting bthreads to NuSMV modules")
        bt_list = []
        for i in range(self.bthread_num):
            bt_list.append(self._bthread_to_module(lambda: self.bprogram_generator().bthreads[i],
                                                   self.bprogram_generator().bthreads[i].__name__ + str(i),
                                                   self.event_list))
        if print_info:
            print("Creating main module")
        main = self._main_module(self.event_list, bt_list)
        if print_info:
            print("Loading model into NuSMV")

        temp_dir = tempfile.TemporaryDirectory()
        with open(temp_dir.name + "/bp_model.smv", "w") as f:
            for bt in bt_list:
                f.write(str(bt))
                f.write("\n")
            f.write(str(main))
            f.write("\n")
            f.write("LTLSPEC " + str(spec).strip())
        pynusmv.glob.load_from_file(temp_dir.name + "/bp_model.smv")

        if type == "BMC":
            if print_info:
                print("Computing model")
            pynusmv.glob.flatten_hierarchy()
            pynusmv.glob.encode_variables()
            pynusmv.glob.build_boolean_model()
            bmc_setup()
            spec = pynusmv.glob.prop_database()[0].expr
            if print_info:
                print("Checking LTL spec")
            with BmcSupport():
                fml = Node.from_ptr(parse_ltl_spec(str(spec).strip()))
                fsm = master_be_fsm()
                problem = ltlspec.generate_ltl_problem(fsm, fml, bound)
                fsm = master_be_fsm()
                cnf = problem.to_cnf(Polarity.POSITIVE)

                solver = SatSolverFactory.create()
                solver += cnf
                solver.polarity(cnf, Polarity.POSITIVE)
                solution = solver.solve()
                result = solution != SatSolverResult.SATISFIABLE

                if not result and find_counterexample:
                    if print_info:
                        print("Finding counterexample")
                    cnt_ex = bmcutils.generate_counter_example(fsm, problem, solver, bound, "Violation")
                    explanation_str = ""
                    first_loop_signal = False
                    for step in cnt_ex:
                        if step.is_loopback:
                            if first_loop_signal:
                                break
                            first_loop_signal = True
                            explanation_str += "-- Loop starts here" + "\n"
                        for symbol, value in step:
                            if str(symbol) == "event":
                                explanation_str += str(value) + "\n"
                else:
                    explanation_str = None
        elif type == "BDD":
            if print_info:
                print("Computing model")
            pynusmv.glob.compute_model()

            # spec = prop.ag(prop.af(prop.atom(("must_finish = FALSE"))))
            spec = pynusmv.glob.prop_database()[0].expr

            if print_info:
                print("Checking LTL spec")
            #fsm = pynusmv.glob.prop_database().master.bddFsm
            result = check_ltl_spec(spec)
            if not result and find_counterexample:
                if print_info:
                    print("Finding counterexample")
                _, explanation = check_explain_ltl_spec(spec)
                explanation_str = ""
                first_loop_signal = False
                for state in explanation[2:-1:2]:
                    if state == explanation[2::2][-1]:
                        if first_loop_signal:
                            break
                        first_loop_signal = True
                        explanation_str += "-- Loop starts here" + "\n"
                    explanation_str += state["event"] + "\n"
            else:
                explanation_str = None
        else:
            pynusmv.init.deinit_nusmv()
            temp_dir.cleanup()
            raise ValueError("Unknown type. Use 'BDD' or 'BMC'")

        pynusmv.init.deinit_nusmv()
        temp_dir.cleanup()
        return result, explanation_str

    def _bthread_to_module(self, bthread_generator, bthread_name, event_list):

        dfs = DFSBThread(bthread_generator, SimpleEventSelectionStrategy(), event_list)
        init_s, visited, requested, blocked = dfs.run(return_requested_and_blocked=True)

        visited = dict([(k, v) for k, v in enumerate(visited)])
        id_to_change = [k for k, v in visited.items() if v == init_s][0]
        s_to_change = visited[0]
        visited[id_to_change] = s_to_change
        visited[0] = init_s
        rev_visited = {v: k for k, v in visited.items()}

        bt1_mod_dict = {}
        bt1_mod_dict["event"] = Identifier("event")
        bt1_mod_dict["ARGS"] = [bt1_mod_dict["event"]]
        bt1_mod_dict["state"] = Var(Range(0, len(visited)))

        bt1_mod_dict.update({
            e.name + "_requested": Var(Boolean(), name=e.name + "_requested") for e in requested
        })
        bt1_mod_dict.update({
            e.name + "_blocked": Var(Boolean(), name=e.name + "_blocked") for e in blocked
        })
        # bt1_mod_dict["must_finish"] = Var(Boolean())
        bt1_mod_dict["INIT"] = [bt1_mod_dict["state"] == 0]
        bt1_mod_dict_assign = {}
        for e in requested:
            case_list = []
            for i, node in visited.items():
                if isinstance(node.data.get(request, {}), BEvent):
                    if e == node.data.get(request, {}):
                        case_list.append((bt1_mod_dict["state"] == i, Trueexp()))
                    else:
                        case_list.append((bt1_mod_dict["state"] == i, Falseexp()))
                else:
                    if e in node.data.get(request, {}):
                        case_list.append((bt1_mod_dict["state"] == i, Trueexp()))
                    else:
                        case_list.append((bt1_mod_dict["state"] == i, Falseexp()))
            case_list = tuple(case_list)
            bt1_mod_dict_assign[e.name + "_requested"] = Case(case_list + ((Trueexp(), Falseexp()),))
        for e in blocked:
            case_list = []
            for i, node in visited.items():
                if isinstance(node.data.get(block, {}), BEvent):
                    if e == node.data.get(block, {}):
                        case_list.append((bt1_mod_dict["state"] == i, Trueexp()))
                    else:
                        case_list.append((bt1_mod_dict["state"] == i, Falseexp()))
                else:
                    if e in node.data.get(block, {}):
                        case_list.append((bt1_mod_dict["state"] == i, Trueexp()))
                    else:
                        case_list.append((bt1_mod_dict["state"] == i, Falseexp()))
            case_list = tuple(case_list)
            bt1_mod_dict_assign[e.name + "_blocked"] = Case(case_list + ((Trueexp(), Falseexp()),))

        case_list = []
        for i, node in visited.items():
            d = {}
            for e, next_node in node.transitions.items():
                d[rev_visited[next_node]] = d.get(rev_visited[next_node], []) + [e]
            for j, events in d.items():
                if i == j:
                    continue  # self loop not necessary
                or_chain = Falseexp()
                for e in events:
                    or_chain = Or(bt1_mod_dict["event"].next() == e.name, or_chain)
                case_list.append((And(bt1_mod_dict["state"] == i, or_chain), j))
        bt1_mod_dict_assign["next(state)"] = Case(tuple(case_list) + ((Trueexp(), bt1_mod_dict["state"]),))
        # true_set = set([i for i, node in visited.items() if node.must_finish])
        # true_or_chain = Falseexp()
        # false_or_chain = Falseexp()
        # for i, node in visited.items():
        #     if i in true_set:
        #         true_or_chain = Or(bt1_mod_dict["state"] == i, true_or_chain)
        #     else:
        #         false_or_chain = Or(bt1_mod_dict["state"] == i, false_or_chain)
        # bt1_mod_dict_assign["must_finish"] = Case(((true_or_chain, Trueexp()),
        #                                            (false_or_chain, Falseexp()),
        #                                            (Trueexp(), Falseexp())))
        bt1_mod_dict["ASSIGN"] = bt1_mod_dict_assign
        return type(bthread_name, (Module,), bt1_mod_dict)

    def _main_module(self, event_list, bt_list):
        mod_dict = {}
        mod_dict["event"] = Var(Scalar(tuple(["BPROGRAM_START", "BPROGRAM_DONE"] + [x.name for x in event_list])))
        bt_modules = []
        for i, bt in enumerate(bt_list):
            mod_dict["bt" + str(i)] = Var(bt(mod_dict["event"]))
            bt_modules.append(mod_dict["bt" + str(i)])
        mod_dict["INIT"] = [mod_dict["event"] == "BPROGRAM_START"]
        mod_dict_define = {}
        for e in event_list:
            mod_dict_define[e.name + "_requested"] = Falseexp()
            mod_dict_define[e.name + "_blocked"] = Falseexp()
        all_requested = set()
        all_blocked = set()
        for i in range(len(bt_list)):
            for v in mod_dict["bt" + str(i)].type.VAR:
                if v.name in mod_dict_define:
                    if v.name.endswith("_requested"):
                        all_requested.add(v.name)
                    elif v.name.endswith("_blocked"):
                        all_blocked.add(v.name)
                    mod_dict_define[v.name] = Or("bt" + str(i) + "." + v.name, mod_dict_define[v.name])
        # mod_dict_define["must_finish"] = Falseexp()
        # for i in range(len(bt_list)):
        #     mod_dict_define["must_finish"] = Or("bt" + str(i) + ".must_finish", mod_dict_define["must_finish"])
        for e in event_list:
            mod_dict_define[e.name + "_enabled"] = And(e.name + "_requested", "!" + e.name + "_blocked")
        mod_dict["DEFINE"] = mod_dict_define
        trans_statement = NotEqual("next(event)", "BPROGRAM_START")
        any_enabled = Falseexp()
        for e in event_list:
            trans_statement = And(trans_statement, Implies("!" + e.name + "_enabled", NotEqual("next(event)", e.name)))
            any_enabled = Or(any_enabled, e.name + "_enabled")
        trans_statement = And(trans_statement, Implies(any_enabled, NotEqual("next(event)", "BPROGRAM_DONE")))
        trans_statement = And(trans_statement, Implies(Equal("event", "BPROGRAM_DONE"), Equal("next(event)", "BPROGRAM_DONE")))
        mod_dict["TRANS"] = [trans_statement]
        return type("main", (Module,), mod_dict)



import gymnasium as gym
from bppy.gym.bp_action_space import BPActionSpace
from bppy.gym.simple_bp_observation_space import SimpleBPObservationSpace
from bppy import SolverBasedEventSelectionStrategy
import warnings
import random

class BPEnv(gym.Env):
    """
    Custom environment that adheres to OpenAI's Gym interface for simulating
    the behavior of a BProgram. This class is useful for integrating BPrograms
    into reinforcement learning algorithms.

    Attributes
    ----------
    action_space : :class:`BPActionSpace <bppy.gym.bp_action_space.BPActionSpace>`
        Defines the space of possible actions, i.e., events.
    observation_space : :class:`BPObservationSpace <bppy.gym.bp_observation_space.BPObservationSpace>`
        Defines the space of possible observations.
    reward_function : function, optional
        A custom function to compute the reward.

    """
    def __init__(self, bprogram_generator, action_list, observation_space=None, action_space=None, reward_function=None):
        """
        Initializes the BPEnv environment.

        Parameters
        ----------
        bprogram_generator : function
            A function that generates a new instance of the BProgram.
        action_list : list
            List of all events defined as actions.
        observation_space : :class:`BPObservationSpace <bppy.gym.bp_observation_space.BPObservationSpace>`, optional
            Space of possible observations. Defaults to
            :class:`SimpleBPObservationSpace <bppy.gym.simple_bp_observation_space.SimpleBPObservationSpace>`.
        reward_function : function, optional
            A custom function to compute the reward. If not provided, the default reward is the sum of all b-thread
            rewards.
        """
        self.metadata = {}
        self.bprogram = None
        self.bprogram_generator = bprogram_generator
        self.action_space = BPActionSpace(action_list)
        self.event_list = action_list
        self.reward_function = reward_function
        if self.reward_function is None:
            self.reward_function = lambda rewards: sum(filter(None, rewards))
        self.observation_space = observation_space
        if self.observation_space is None:
            self.observation_space = SimpleBPObservationSpace(self.bprogram_generator, action_list)

    def step(self, action):
        """
        Executes the given action and returns the resulting observation, reward, done flag, and additional information.

        Parameters
        ----------
        action : int
            An index representing the event to be executed.

        Returns
        -------
        observation : object
            The state of the environment after executing the action.
        reward : float
            The reward obtained by executing the action.
        done : bool
            Whether the episode has ended.
        truncated : bool
            Not used for this environment.
        info : dict
            Additional information for debugging.
        """
        if self.bprogram is None:
            raise RuntimeError("You must call reset() before calling step()")
        if not self.action_space.contains(action):
            return self._state(), 0, True, None, {"message": "Last event is disabled"}
        event = self.action_space.event_list[action]
        self.bprogram.advance_bthreads(self.bprogram.tickets, event)
        while not self._step_done():
            events_are_actions = [e in self.event_list for e in
                                  self.bprogram.event_selection_strategy.selectable_events(self.bprogram.tickets)]
            if any(events_are_actions):
                selectable_events = self.bprogram.event_selection_strategy.selectable_events(self.bprogram.tickets)
                if selectable_events:
                    self.bprogram.advance_bthreads(self.bprogram.tickets, random.choice(tuple(selectable_events)))
            else:
                self.bprogram.advance_bthreads(self.bprogram.tickets,
                                               self.bprogram.event_selection_strategy.select(self.bprogram.tickets))
        done = self.bprogram.event_selection_strategy.selectable_events(self.bprogram.tickets).__len__() == 0
        return self._state(), self._reward(), done, None, {}

    def reset(self, seed=None, options=None):
        """
        Resets the environment to its initial state.

        Parameters
        ----------
        seed : int, optional
            A seed for the random number generator.
        options : dict, optional
            Additional options for resetting the environment.

        Returns
        -------
        observation : object
            The initial state of the environment.
        info : dict
            Not used for this environment.
        """
        super().reset(seed=seed, options=options)
        self.bprogram = self.bprogram_generator()
        # if isinstance(self.bprogram.event_selection_strategy, SolverBasedEventSelectionStrategy):
        #     raise NotImplementedError("SolverBasedEventSelectionStrategy is currently not supported")
        self.action_space.bprogram = self.bprogram
        self.bprogram.setup()
        while not self._step_done():
            events_are_actions = [e in self.event_list for e in
                                  self.bprogram.event_selection_strategy.selectable_events(self.bprogram.tickets)]
            if any(events_are_actions):
                selectable_events = self.bprogram.event_selection_strategy.selectable_events(self.bprogram.tickets)
                if selectable_events:
                    self.bprogram.advance_bthreads(self.bprogram.tickets, random.choice(tuple(selectable_events)))
            else:
                self.bprogram.advance_bthreads(self.bprogram.tickets,
                                               self.bprogram.event_selection_strategy.select(self.bprogram.tickets))
        return self.observation_space.bp_state_to_gym_space(self._bthreads_states()), {}

    def render(self, mode="human"):
        """
        Not implemented for this environment.
        """
        raise NotImplementedError()

    def close(self):
        """
        Closes the environment, releasing any resources.
        """
        self.bprogram = None

    def get_state(self):
        """
        Returns the current state of the environment.
        """
        return self._state()

    def _bthreads_states(self):
        return [dict([k, v] for k, v in statement.items() if k != "bt") for statement in self.bprogram.tickets]

    def _state(self):
        return self.observation_space.bp_state_to_gym_space(self._bthreads_states())

    def _bthreads_rewards(self):
        return [x.get("localReward") for x in self.bprogram.tickets]

    def _reward(self):
        return self.reward_function(self._bthreads_rewards())

    def _step_done(self):
        events_are_actions = [e in self.event_list for e in self.bprogram.event_selection_strategy.selectable_events(self.bprogram.tickets)]
        if all(events_are_actions):
            return True
        else:
            if any(events_are_actions):
                warnings.warn("Some selectable events are defined in the actions list, while others are not. "
                              "The program will continue running non-action events until all selectable events "
                              "are defined in the actions list.")
            return False



import random
import itertools
from math import prod
from bppy.utils.weighted_sampling import *
from uuid import uuid4

request = "request"
waitFor = "waitFor"
block = "block"
mustFinish = "mustFinish"
priority = "priority"
localReward = "localReward"

class sync(dict):
	def __init__(self, *, request=None, waitFor=None, block=None, mustFinish=None, priority=None, localReward=None, **kwargs):
		#'''TODO: warn if req/waitFor/block are not BEvent or choice'''
		if request is not None:
			self["request"] = request
		if waitFor is not None:
			self["waitFor"] = waitFor
		if block is not None:
			self["block"] = block
		if mustFinish is not None:
			self["mustFinish"] = mustFinish
		if priority is not None:
			self["priority"] = priority
		if localReward is not None:
			self["localReward"] = localReward
		super().__init__(kwargs)


class choice(dict):
	"""
	A class to represent a discrete choice object.

	Keys correspond to the possible choices, and values correspond to the probability of each choice.
	"""
	def __init__(self, data, repeat=1, replace=True, sorted=False, 
				*args, **kwargs):
		if not isinstance(data, dict):
			raise TypeError("data must be a dict")
		if not replace and repeat > len(data):
			raise ValueError("repeat must be smaller than the number of choices")
		self.repeat = repeat
		self.replace = replace
		self.sorted = sorted
		super().__init__(data, **kwargs)
		self._id = uuid4()

	# returns iterable
	def options(self):
		dist = [self.keys(), self.values()]
		if self.replace and not self.sorted:
			pv, pp = [itertools.product(l,
			repeat=self.repeat) for l in dist]
			combined_probs = [prod(event_probs) for event_probs in pp]
		elif self.replace and self.sorted:
			pv, pp = [itertools.combinations_with_replacement(l,
			r=self.repeat) for l in dist]
		elif not self.replace and self.sorted:
			pv = itertools.combinations(dist[0], r=self.repeat)
			pv, combined_probs = zip(*[sequence_probability_nr_s(self, perm)
							 	for perm in pv])
		elif not self.replace and not self.sorted:
			pv, pp = [itertools.permutations(l,
			r=self.repeat) for l in dist]
		else:
			raise RuntimeError("Invalid combination of replace and sorted")
		if self.repeat == 1:
			return zip([v[0] for v in pv], combined_probs)
		return zip(pv, combined_probs)

	def sample(self):
		if self.replace:
			res = random.choices(list(self.keys()),
					self.values(), k=self.repeat)
			if self.repeat == 1:
				return res[0]
			if self.sorted:
				sort(res)
			return res
		else:
			res = weighted_sample_without_replacement(list(self.keys()),
					self.values(), k=self.repeat)
			if self.repeat == 1:
				return res[0]
			if self.sorted:
				res.sort()
			return res
	
	def __eq__(self, other):
		return isinstance(other, choice) and self._id == other._id
